# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from .hyperparameters.augmentation import (
    AugmentationParameters,
    CenterCrop,
    ColorJitter,
    GaussianBlur,
    RandomAffine,
    RandomHorizontalFlip,
    RandomIOUCrop,
    RandomResizeCrop,
    RandomVerticalFlip,
    Tiling,
)
from .hyperparameters.hyperparameters import (
    DatasetPreparationParameters,
    EarlyStopping,
    EvaluationParameters,
    Hyperparameters,
    TrainingHyperParameters,
)
from .project_configuration import (
    AutoTrainingParameters,
    ProjectConfiguration,
    TaskConfig,
    TrainConstraints,
    TrainingParameters,
)
from .training_configuration import (
    Filtering,
    GlobalDatasetPreparationParameters,
    GlobalParameters,
    MaxAnnotationObjects,
    MaxAnnotationPixels,
    MinAnnotationObjects,
    MinAnnotationPixels,
    SubsetSplit,
    TrainingConfiguration,
)

__all__ = [
    "TrainConstraints",
    "AutoTrainingParameters",
    "TrainingParameters",
    "TaskConfig",
    "ProjectConfiguration",
    "SubsetSplit",
    "MinAnnotationPixels",
    "MaxAnnotationPixels",
    "MinAnnotationObjects",
    "MaxAnnotationObjects",
    "Filtering",
    "GlobalDatasetPreparationParameters",
    "GlobalParameters",
    "TrainingConfiguration",
    "CenterCrop",
    "RandomResizeCrop",
    "RandomAffine",
    "RandomHorizontalFlip",
    "RandomVerticalFlip",
    "RandomIOUCrop",
    "ColorJitter",
    "GaussianBlur",
    "Tiling",
    "AugmentationParameters",
    "DatasetPreparationParameters",
    "EarlyStopping",
    "TrainingHyperParameters",
    "EvaluationParameters",
    "Hyperparameters",
]



# INTEL CONFIDENTIAL
#
# Copyright (C) 2024 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were provided to
# you ("License"). Unless the License provides otherwise, you may not use, modify, copy,
# publish, distribute, disclose or transmit this software or the related documents
# without Intel's prior written permission.
#
# This software and the related documents are provided as is,
# with no express or implied warranties, other than those that are expressly stated
# in the License.

"""Module implements the InferenceResultsToPredictionConverter class."""

import abc
import logging
from collections import defaultdict
from typing import Any, NamedTuple

import cv2
import numpy as np
from model_api.models import (
    AnomalyResult,
    ClassificationResult,
    DetectedKeypoints,
    DetectionResult,
    ImageModel,
    ImageResultWithSoftPrediction,
    InstanceSegmentationResult,
    SegmentationModel,
)

from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.containers import LabelList
from geti_sdk.data_models.enums.domain import Domain
from geti_sdk.data_models.label import Label, ScoredLabel
from geti_sdk.data_models.predictions import Prediction
from geti_sdk.data_models.shapes import (
    Ellipse,
    Keypoint,
    Point,
    Polygon,
    Rectangle,
    RotatedRectangle,
)
from geti_sdk.deployment.predictions_postprocessing.utils.segmentation_utils import (
    create_annotation_from_segmentation_map,
)


class InferenceResultsToPredictionConverter(metaclass=abc.ABCMeta):
    """Interface for the converter"""

    def __init__(self, labels: LabelList, configuration: dict[str, Any]):
        self.labels = labels.get_non_empty_labels()
        model_api_labels = configuration["labels"]
        label_ids = configuration.get("label_ids", [])  # default to empty list
        # configuration["labels"] can be a single string or a list of strings
        model_api_labels = (
            model_api_labels.split()  # space separated string
            if isinstance(model_api_labels, str)
            else [str(name) for name in model_api_labels]
        )
        # Create a mapping of label ID to label objects
        self.label_map_ids = {}
        # Legacy OTX (<2.0) model configuration contains label names (without spaces) instead of IDs
        self.legacy_label_map_names = defaultdict(list)

        # get the first empty label
        self.empty_label = labels.get_empty_label()

        for i, label in enumerate(labels):
            self.label_map_ids[str(label.id)] = label
            # Using a dict of list to handle duplicates label names (e.g. "foo bar", "foo_bar")
            self.legacy_label_map_names[label.name.replace(" ", "_")].append(label)
            self.legacy_label_map_names[label.name].append(label)
        self.legacy_label_map_names["otx_empty_lbl"] = [self.empty_label]

        # Create a mapping of ModelAPI label indices to label objects
        self.idx_to_label = {}
        self.str_to_label = {}
        self.model_api_label_map_counts: dict[str, int] = defaultdict(int)

        # Validate that label_ids and model_api_labels are properly matched
        if (n_missing_ids := len(model_api_labels) - len(label_ids)) > 0:
            logging.warning(
                f"Mismatch between label_ids (len={len(label_ids)}) and model_api_labels (len={len(model_api_labels)})."
                f" Using placeholder label IDs for the missing {n_missing_ids} labels."
            )
            for i in range(n_missing_ids):
                label_ids.append(f"generated_label_{i}")

        # Assumes configuration['label_ids'] and configuration['labels'] have the same ordering
        for i, (label_id_str, label_str) in enumerate(zip(label_ids, model_api_labels)):
            try:
                label = self.__get_label(label_str, pos_idx=self.model_api_label_map_counts[label_str])
            except ValueError:
                if label_id_str in self.label_map_ids:
                    # Get the label by its ID in case it has been renamed
                    label = self.label_map_ids[label_id_str]
                    logging.warning(f"Label '{label_str}' has been renamed to '{label.name}'.")
                else:
                    logging.warning(f"Label '{label_str}' cannot be found. It may have been removed.")
            self.idx_to_label[i] = label
            self.str_to_label[label_str] = label
            self.model_api_label_map_counts[label_str] += 1
        logging.info(f"Converter loaded labels with following indices: {self.idx_to_label}")

    def __get_label(self, label_str: str, pos_idx: int) -> Label:
        if label_str in self.label_map_ids:
            return self.label_map_ids[label_str]
        matched_legacy_labels = self.legacy_label_map_names[label_str]
        if pos_idx < len(matched_legacy_labels):
            return matched_legacy_labels[pos_idx]
        raise ValueError(f"Label '{label_str}' (pos_idx={pos_idx}) not found in the label schema")

    def get_label_by_idx(self, label_idx: int) -> Label:
        """
        Get a Label object by its index. It is useful for converting ModelAPI results to Prediction.

        :param label_idx: index of the label from prediction results
        :return: Label corresponding to the index
        """
        return self.idx_to_label[label_idx]

    def get_label_by_str(self, label_str: int) -> Label:
        """
        Get a Label object by its string representation. It is useful for converting ModelAPI results to Prediction.

        :param label_str: string representation of the label from prediction results
        :return: Label corresponding to the string
        """
        return self.str_to_label[label_str]

    @abc.abstractmethod
    def convert_to_prediction(self, inference_results: NamedTuple, **kwargs) -> Prediction:
        """
        Convert raw inference results to the Prediction format.

        :param inference_results: raw predictions from inference
        :return: Prediction object containing the shapes obtained from the raw predictions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def convert_saliency_map(self, inference_results: NamedTuple, **kwargs) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: raw predictions from inference
        :return: Prediction object containing the shapes obtained from the raw predictions.
        """
        raise NotImplementedError


class ClassificationToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Classification predictions to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: configuration dictionary containing additional
        parameters
    """

    def __init__(self, labels: LabelList, configuration: dict[str, Any]):
        super().__init__(labels, configuration)

    def convert_to_prediction(
        self,
        inference_results: ClassificationResult,
        image_shape: tuple[int, int, int],
        **kwargs,
    ) -> Prediction:  # noqa: ARG003
        """
        Convert ModelAPI ClassificationResult inference results to Prediction object.

        :param inference_results: classification labels represented in ModelAPI format
            (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        labels = []
        for label in inference_results.top_labels:
            label_idx, label_name, label_prob = label
            scored_label = ScoredLabel.from_label(label=self.get_label_by_idx(label_idx), probability=label_prob)
            labels.append(scored_label)

        if not labels and self.empty_label:
            labels = [ScoredLabel.from_label(self.empty_label, probability=0)]

        annotations = Annotation(
            shape=Rectangle.generate_full_box(image_shape[1], image_shape[0]),
            labels=labels,
        )
        return Prediction([annotations])

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: tuple[int, int, int],
    ) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format
            (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        saliency_map = inference_results.saliency_map
        if len(saliency_map) == 0:
            return None
        saliency_map = cv2.resize(
            np.transpose(saliency_map.squeeze(0), (1, 2, 0)),
            dsize=(image_shape[1], image_shape[0]),
            interpolation=cv2.INTER_CUBIC,
        )
        if len(saliency_map.shape) == 2:
            saliency_map = np.expand_dims(saliency_map, axis=-1)
        saliency_map = np.transpose(saliency_map, (2, 0, 1))  # shape: (N classes, h, w)
        return {label.name: saliency_map[i] for i, label in enumerate(self.labels.get_non_empty_labels())}

    def _get_label_by_prediction_name(self, name: str) -> Label:
        """
        Get a Label object by its predicted name.

        :param name: predicted name of the label
        :return: Label corresponding to the name
        :raises KeyError: if the label is not found in the LabelList
        """
        try:
            return self.labels.get_by_name(name=name)
        except KeyError:
            # If the label is not found, we try to find it by legacy name (replacing spaces with underscores)
            for label in self.labels:
                legacy_name = label.name.replace(" ", "_")
                if legacy_name == name:
                    logging.warning(f"Found label `{label.name}` using its legacy name `{legacy_name}`.")
                    return label
        raise KeyError(f"Label named `{name}` was not found in the LabelList")


class DetectionToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Detection objects to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def __init__(self, labels: LabelList, configuration: dict[str, Any]):
        super().__init__(labels, configuration)
        self.use_ellipse_shapes = False
        self.confidence_threshold = 0.0
        if "use_ellipse_shapes" in configuration:
            self.use_ellipse_shapes = configuration["use_ellipse_shapes"]
        if "confidence_threshold" in configuration:
            self.confidence_threshold = configuration["confidence_threshold"]

    def _detection2array(self, detection: DetectionResult) -> np.ndarray:
        """
        Convert DetectionResult to a numpy array.

        :param detection: list of OpenVINO Detection containing [score, id, xmin, ymin, xmax, ymax]
        :return: numpy array with [label, confidence, x1, y1, x2, y2]
        """
        # Filter valid detections first
        valid_detections = [
            (score, label, bbox)
            for score, label, bbox in zip(detection.scores, detection.labels, detection.bboxes)
            if (bbox[2] - bbox[0]) * (bbox[3] - bbox[1]) >= 1.0
        ]

        if not valid_detections:
            return np.empty((0, 6), dtype=np.float32)

        # Pre-allocate arrays with known size
        n_detections = len(valid_detections)
        result = np.empty((n_detections, 6), dtype=np.float32)

        for i, (score, label, bbox) in enumerate(valid_detections):
            result[i] = [label, score, bbox[0], bbox[1], bbox[2], bbox[3]]

        return result

    def convert_to_prediction(self, inference_results: DetectionResult, **kwargs) -> Prediction:
        """
        Convert ModelAPI DetectionResult inference results to Prediction object.

        :param inference_results: detection represented in ModelAPI format (label, confidence, x1, y1, x2, y2).

        _Note:
            - `label` can be any integer that can be mapped to `self.labels`
            - `confidence` should be a value between 0 and 1
            - `x1`, `x2`, `y1` and `y2` are expected to be in pixel
        :return: Prediction object containing the boxes obtained from the prediction
        """
        detections = self._detection2array(inference_results)

        annotations = []
        if len(detections) and detections.shape[1:] < (6,) or detections.shape[1:] > (7,):
            raise ValueError(
                f"Shape of prediction is not expected, expected (n, 7) or (n, 6) but got {detections.shape}"
            )

        for detection in detections:
            # Some OpenVINO models use an output shape of [7,]
            # If this is the case, skip the first value as it is not used
            _detection = detection[1:] if detection.shape == (7,) else detection

            label_index = int(_detection[0])
            confidence = _detection[1]
            scored_label = ScoredLabel.from_label(self.get_label_by_idx(label_index), confidence)
            coords = _detection[2:]
            shape: Ellipse | Rectangle

            if confidence < self.confidence_threshold:
                continue

            bbox_width = coords[2] - coords[0]
            bbox_height = coords[3] - coords[1]
            if self.use_ellipse_shapes:
                shape = Ellipse(coords[0], coords[1], bbox_width, bbox_height)
            else:
                shape = Rectangle(coords[0], coords[1], bbox_width, bbox_height)

            annotation = Annotation(shape=shape, labels=[scored_label])
            annotations.append(annotation)
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: tuple[int, int, int],
    ) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format
            (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        saliency_map = inference_results.saliency_map
        if len(saliency_map) == 0:
            return None
        if isinstance(saliency_map, list):
            saliency_map = np.array(
                [smap if len(smap) > 0 else np.zeros(image_shape[:2], dtype=np.uint8) for smap in saliency_map]
            )
        elif isinstance(saliency_map, np.ndarray):
            saliency_map = saliency_map.squeeze(0)
        else:
            raise ValueError(f"Unsupported saliency map type: {type(saliency_map)}. Expected list or numpy array.")
        saliency_map = cv2.resize(
            np.transpose(saliency_map, (1, 2, 0)),
            dsize=(image_shape[1], image_shape[0]),
            interpolation=cv2.INTER_CUBIC,
        )
        if len(saliency_map.shape) == 2:
            saliency_map = np.expand_dims(saliency_map, axis=-1)
        saliency_map = np.transpose(saliency_map, (2, 0, 1))  # shape: (N classes, h, w)
        return {label.name: saliency_map[i] for i, label in enumerate(self.labels)}


class RotatedRectToPredictionConverter(DetectionToPredictionConverter):
    """
    Converts ModelAPI Rotated Detection objects to Prediction.
    """

    def convert_to_prediction(self, inference_results: InstanceSegmentationResult, **kwargs) -> Prediction:
        """
        Convert ModelAPI instance segmentation inference results to a rotated bounding box annotation format.

        :param inference_results: segmentation represented in ModelAPI format
        :return: Prediction object containing the rotated boxes obtained from the segmentation contours
        :raises ValueError: if metadata is missing from the preprocess step
        """
        annotations = []
        shape: RotatedRectangle | Ellipse
        for bbox, label, mask, score in zip(
            inference_results.bboxes,
            inference_results.labels,
            inference_results.masks,
            inference_results.scores,
        ):
            label = self.get_label_by_idx(label)
            if score < self.confidence_threshold or label.is_empty:
                continue
            if self.use_ellipse_shapes:
                shape = Ellipse(bbox[0], bbox[1], bbox[2] - bbox[0], bbox[3] - bbox[1])
                annotations.append(
                    Annotation(
                        shape=shape,
                        labels=[ScoredLabel.from_label(label, float(score))],
                    )
                )
            else:
                mask = mask.astype(np.uint8)
                contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                        continue
                    points = [
                        Point(
                            x=point[0],
                            y=point[1],
                        )
                        for point in cv2.boxPoints(cv2.minAreaRect(contour))
                    ]
                    shape = Polygon(points=points)
                    annotations.append(
                        Annotation(
                            shape=RotatedRectangle.from_polygon(shape),
                            labels=[ScoredLabel.from_label(label, float(score))],
                        )
                    )
        return Prediction(annotations)


class MaskToAnnotationConverter(DetectionToPredictionConverter):
    """
    Converts DetectionBox Predictions ModelAPI to Prediction object.
    """

    def convert_to_prediction(self, inference_results: Any, **kwargs: dict[str, Any]) -> Prediction:
        """
        Convert inference results to Prediction object.

        :param inference_results: Raw inference results from the model.
        :return: Prediction object.
        """
        annotations = []
        shape: Polygon | Ellipse
        for obj in inference_results.segmentedObjects:
            if obj.score < self.confidence_threshold:
                continue
            if self.use_ellipse_shapes:
                shape = Ellipse(obj.xmin, obj.ymin, obj.xmax - obj.xmin, obj.ymax - obj.ymin)
                annotations.append(
                    Annotation(
                        shape=shape,
                        labels=[ScoredLabel.from_label(self.get_label_by_idx(obj.id), float(obj.score))],
                    )
                )
            else:
                mask = obj.mask.astype(np.uint8)
                contours, hierarchies = cv2.findContours(mask, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_SIMPLE)
                if hierarchies is None:
                    continue
                for contour, hierarchy in zip(contours, hierarchies[0]):
                    if hierarchy[3] != -1:
                        continue
                    if len(contour) <= 2 or cv2.contourArea(contour) < 1.0:
                        continue
                    contour = list(contour)
                    points = [
                        Point(
                            x=point[0][0],
                            y=point[0][1],
                        )
                        for point in contour
                    ]
                    shape = Polygon(points=points)
                    annotations.append(
                        Annotation(
                            shape=shape,
                            labels=[ScoredLabel.from_label(self.get_label_by_idx(obj.id), float(obj.score))],
                        )
                    )
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: tuple[int, int, int],
    ) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format
            (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        if len(inference_results.saliency_map) == 0:
            return None
        # Model API returns a list of np.ndarray for each label
        # Including `no_object` which is empty
        saliency_map = np.array(
            [
                smap if len(smap) > 0 else np.zeros(image_shape[:2], dtype=np.uint8)
                for smap in inference_results.saliency_map
            ]
        )  # shape: (N classes, h, w)
        return {label.name: saliency_map[i] for i, label in enumerate(self.labels)}


class SegmentationToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Segmentation objects to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: model configuration setting
    :param model: SegmentationModel instance, needed for getting contours
    """

    def __init__(self, labels: LabelList, configuration: dict[str, Any], model: SegmentationModel):
        super().__init__(labels, configuration)
        self.model = model

    def get_label_by_idx(self, label_idx: int) -> Label:
        """
        Get a Label object by its index. It is useful for converting ModelAPI results to Prediction.

        # NB: For segmentation results, index=0 is reserved for the background label

        :param label_idx: index of the label from prediction results
        :return: Label corresponding to the index
        """
        self.idx_to_label[-1] = self.empty_label
        return super().get_label_by_idx(label_idx - 1)

    def convert_to_prediction(
        self,
        inference_results: ImageResultWithSoftPrediction,
        **kwargs,  # noqa: ARG002
    ) -> Prediction:
        """
        Convert ModelAPI instance segmentation inference results to Prediction object.

        :param inference_results: segmentation represented in ModelAPI format
        :return: Prediction object containing the contour polygon obtained from the segmentation
        """
        contours = self.model.get_contours(inference_results)

        annotations: list[Annotation] = []
        for contour in contours:
            label = self.get_label_by_str(contour.label)
            if len(contour.shape) > 0 and not label.is_empty:
                approx_curve = cv2.approxPolyDP(contour.shape, 1.0, True)
                if len(approx_curve) > 2:
                    points = [Point(x=p[0][0], y=p[0][1]) for p in contour.shape]
                    annotations.append(
                        Annotation(
                            shape=Polygon(points=points),
                            labels=[ScoredLabel.from_label(label=label, probability=contour.probability)],
                        )
                    )
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: tuple[int, int, int],
    ) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format
            (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        saliency_map = inference_results.saliency_map
        if len(saliency_map) == 0:
            return None
        saliency_map = np.transpose(saliency_map, (2, 0, 1))  # shape: (N classes, h, w)
        return {label.name: saliency_map[i] for i, label in self.idx_to_label.items() if not label.is_empty}


class AnomalyToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Convert ModelAPI AnomalyResult predictions to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: model configuration setting
    """

    def __init__(self, labels: LabelList, configuration: dict[str, Any]):
        self.normal_label = next(label for label in labels if not label.is_anomalous)
        self.anomalous_label = next(label for label in labels if label.is_anomalous)
        if configuration is not None and "domain" in configuration:
            self.domain = configuration["domain"]

    def convert_to_prediction(self, inference_results: AnomalyResult, image_shape: tuple[int], **kwargs) -> Prediction:  # noqa: ARG002
        """
        Convert ModelAPI AnomalyResult inferenceresults to sc_sdk annotations.

        :param inference_results: anomaly result represented in ModelAPI format (same for all anomaly tasks)
        :return: Prediction object based on the specific anomaly task:
            - Classification: single label (normal or anomalous).
            - Segmentation: contour polygon representing the segmentation.
            - Detection: predicted bounding boxes.
        """
        pred_label = inference_results.pred_label
        label = self.anomalous_label if pred_label.lower() in ("anomaly", "anomalous") else self.normal_label
        annotations: list[Annotation] = []
        if self.domain == Domain.ANOMALY_CLASSIFICATION or self.domain == Domain.ANOMALY:
            scored_label = ScoredLabel.from_label(label=label, probability=float(inference_results.pred_score))
            annotations = [
                Annotation(
                    shape=Rectangle.generate_full_box(*image_shape[1::-1]),
                    labels=[scored_label],
                )
            ]
        elif self.domain == Domain.ANOMALY_SEGMENTATION:
            annotations = create_annotation_from_segmentation_map(
                hard_prediction=inference_results.pred_mask,
                soft_prediction=inference_results.anomaly_map.squeeze(),
                label_map={0: self.normal_label, 1: self.anomalous_label},
            )
        elif self.domain == Domain.ANOMALY_DETECTION:
            for box in inference_results.pred_boxes:
                annotations.append(
                    Annotation(
                        shape=Rectangle(box[0], box[1], box[2] - box[0], box[3] - box[1]),
                        labels=[
                            ScoredLabel.from_label(
                                label=self.anomalous_label,
                                probability=inference_results.pred_score,
                            )
                        ],
                    )
                )
        else:
            raise ValueError(
                f"Cannot convert inference results for task '{self.domain.name}'. Only Anomaly tasks are supported."
            )
        if not annotations:
            scored_label = ScoredLabel.from_label(label=self.normal_label, probability=0)
            annotations = [
                Annotation(
                    labels=[scored_label],
                    shape=Rectangle.generate_full_box(*image_shape[1::-1]),
                )
            ]
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: tuple[int, int, int],
    ) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: classification labels represented in ModelAPI format
            (label_index, label_name, confidence)
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        # Normalizing Anomaly map
        saliency_map = inference_results.anomaly_map
        saliency_map -= saliency_map.min()
        saliency_map = saliency_map / (saliency_map.max() + 1e-12) * 255
        saliency_map = np.round(saliency_map).astype(np.uint8)  # shape: (h, w)
        return {self.anomalous_label.name: saliency_map}


class KeypointDetectionToPredictionConverter(InferenceResultsToPredictionConverter):
    """
    Converts ModelAPI Keypoint Detection objects to Prediction object.

    :param labels: LabelList containing the label info of the task
    :param configuration: optional model configuration setting
    """

    def __init__(self, labels: LabelList, configuration: dict[str, Any]):
        super().__init__(labels, configuration)

    def convert_to_prediction(self, inference_results: DetectedKeypoints, **kwargs) -> Prediction:
        """
        Convert ModelAPI KeypointDetectionResult inference results to Prediction object.

        :param inference_results: keypoints represented in ModelAPI format (keypoints, scores).

        _Note:
            - `keypoints` a list of keypoint pairs expected to be in pixel e.g. [[x1, y1], [x2, y2], ...]
            - `scores` should be a value between 0 and 1
        :return: Prediction object containing the keypoints obtained from the prediction
        """
        annotations = []
        for label_idx, keypoint_score in enumerate(zip(inference_results.keypoints, inference_results.scores)):
            shape = Keypoint(x=keypoint_score[0][0], y=keypoint_score[0][1], is_visible=True)
            label = self.get_label_by_idx(label_idx=label_idx)
            scored_label = ScoredLabel.from_label(label=label, probability=keypoint_score[1])
            annotation = Annotation(shape=shape, labels=[scored_label])
            annotations.append(annotation)
        return Prediction(annotations)

    def convert_saliency_map(
        self,
        inference_results: NamedTuple,
        image_shape: tuple[int, int, int],
    ) -> dict[str, np.ndarray] | None:
        """
        Extract a saliency map from inference results and return in a unified format.

        :param inference_results: keypoints represented in ModelAPI format (keypoints, scores).
        :param image_shape: shape of the input image
        :return: Prediction object with corresponding label
        """
        raise NotImplementedError


class ConverterFactory:
    """
    Factory class for creating inference result to prediction converters based on the model's task.
    """

    @staticmethod
    def create_converter(
        labels: LabelList,
        domain: Domain,
        configuration: dict[str, Any],
        model: ImageModel,
    ) -> InferenceResultsToPredictionConverter:
        """
        Create the appropriate inference converter object according to the model's task.

        :param labels: The labels of the model
        :param domain: The domain to which the converter applies
        :param configuration: configuration for the converter
        :param model: ImageModel instance
        :return: The created inference result to prediction converter.
        :raises ValueError: If the task type cannot be determined from the label schema.
        """
        if domain == Domain.CLASSIFICATION:
            return ClassificationToPredictionConverter(labels, configuration)
        if domain == Domain.DETECTION:
            return DetectionToPredictionConverter(labels, configuration)
        if domain == Domain.SEGMENTATION:
            return SegmentationToPredictionConverter(labels, configuration, model=model)
        if domain == Domain.ROTATED_DETECTION:
            return RotatedRectToPredictionConverter(labels, configuration)
        if domain == Domain.INSTANCE_SEGMENTATION:
            return MaskToAnnotationConverter(labels, configuration)
        if domain == Domain.KEYPOINT_DETECTION:
            return KeypointDetectionToPredictionConverter(labels, configuration)
        if domain in (
            Domain.ANOMALY_CLASSIFICATION,
            Domain.ANOMALY_SEGMENTATION,
            Domain.ANOMALY_DETECTION,
            Domain.ANOMALY,
        ):
            configuration.update({"domain": domain})
            return AnomalyToPredictionConverter(labels, configuration)
        raise ValueError(f"Cannot create inferencer for task type '{domain.name}'.")



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from pydantic import BaseModel, Field


class CenterCrop(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable center crop",
        description="Whether to apply center cropping to the image",
    )
    ratio: float = Field(
        gt=0.0,
        default=1.0,
        title="Crop ratio",
        description="Ratio of original dimensions to keep when cropping",
    )


class RandomResizeCrop(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable random resize crop",
        description="Whether to apply random resize and crop to the image",
    )
    ratio: float = Field(
        gt=0.0,
        default=1.0,
        title="Crop resize ratio",
        description="Ratio of original dimensions to apply during resize crop operation",
    )


class RandomAffine(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable random affine",
        description="Whether to apply random affine transformations to the image",
    )
    degrees: float = Field(
        ge=0.0,
        default=0.0,
        title="Rotation degrees",
        description="Maximum rotation angle in degrees",
    )
    translate_x: float = Field(
        default=0.0,
        title="Horizontal translation",
        description="Maximum horizontal translation as a fraction of image width",
    )
    translate_y: float = Field(
        default=0.0,
        title="Vertical translation",
        description="Maximum vertical translation as a fraction of image height",
    )
    scale: float = Field(
        default=1.0,
        title="Scale factor",
        description="Scaling factor for the image during affine transformation",
    )


class RandomHorizontalFlip(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable random horizontal flip",
        description="Whether to apply random flip images horizontally along the vertical axis (swap left and right)",
    )


class RandomVerticalFlip(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable random vertical flip",
        description="Whether to apply random flip images vertically along the horizontal axis (swap top and bottom)",
    )


class RandomIOUCrop(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable random IoU crop",
        description="Whether to apply random cropping based on IoU criteria",
    )


class ColorJitter(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable color jitter",
        description="Whether to apply random color jitter to the image",
    )


class GaussianBlur(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable Gaussian blur",
        description="Whether to apply Gaussian blur to the image",
    )
    kernel_size: int = Field(gt=0, default=3, title="Kernel size", description="Size of the Gaussian kernel")


class Tiling(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable tiling",
        description="Whether to apply tiling to the image",
    )
    adaptive_tiling: bool = Field(
        default=False,
        title="Adaptive tiling",
        description="Whether to use adaptive tiling based on image content",
    )
    tile_size: int = Field(gt=0, default=128, title="Tile size", description="Size of each tile in pixels")
    tile_overlap: float = Field(
        ge=0.0,
        lt=1.0,
        default=0.5,
        title="Tile overlap",
        description="Overlap between adjacent tiles as a fraction of tile size",
    )


class AugmentationParameters(BaseModel):
    """Configuration parameters for data augmentation during training."""

    center_crop: CenterCrop | None = Field(
        default=None,
        title="Center crop",
        description="Settings for center cropping images",
    )
    random_resize_crop: RandomResizeCrop | None = Field(
        default=None,
        title="Random resize crop",
        description="Settings for random resize and crop augmentation",
    )
    random_affine: RandomAffine | None = Field(
        default=None,
        title="Random affine",
        description="Settings for random affine transformations",
    )
    random_horizontal_flip: RandomHorizontalFlip | None = Field(
        default=None,
        title="Random horizontal flip",
        description="Randomly flip images horizontally along the vertical axis (swap left and right)",
    )
    random_vertical_flip: RandomVerticalFlip | None = Field(
        default=None,
        title="Random vertical flip",
        description="Randomly flip images vertically along the horizontal axis (swap top and bottom)",
    )
    random_iou_crop: RandomIOUCrop | None = Field(
        default=None,
        title="Random IoU crop",
        description="Randomly crop images based on Intersection over Union (IoU) criteria",
    )
    color_jitter: ColorJitter | None = Field(
        default=None,
        title="Color jitter",
        description="Settings for random color jitter (brightness, contrast, saturation, hue)",
    )
    gaussian_blur: GaussianBlur | None = Field(
        default=None,
        title="Gaussian blur",
        description="Settings for Gaussian blur augmentation",
    )
    tiling: Tiling | None = Field(default=None, title="Tiling", description="Settings for image tiling")



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import re
from collections.abc import Callable
from importlib import resources
from typing import Any

import numpy as np
from pathvalidate import sanitize_filepath

from geti_sdk.data_models import (
    Annotation,
    Label,
    OptimizedModel,
    Prediction,
    Project,
    ScoredLabel,
    Task,
    TaskType,
)
from geti_sdk.data_models.shapes import Rectangle
from geti_sdk.deployment.data_models import ROI, IntermediateInferenceResult

try:
    OVMS_README_PATH = str(resources.files("geti_sdk.deployment.resources") / "OVMS_README.md")
except AttributeError:
    with resources.path("geti_sdk.deployment.resources", "OVMS_README.md") as data_path:
        OVMS_README_PATH = str(data_path)


def generate_ovms_model_name(project: Project, model: OptimizedModel, omit_version: bool = True) -> str:
    """
    Generate a valid name for a model to be uploaded to the OpenVINO Model Server.

    :param project: Project the model belongs to
    :param model: The model for which to generate the name
    :param omit_version: True not to include the version of the model in the name.
        This is useful if the name needs to be used to create a directory structure,
        since including the version will break this.
    :return: String containing the name of the model
    """
    model_name = sanitize_filepath(project.name + "_" + model.name)
    model_name = model_name.replace(" ", "_")
    model_name = model_name.replace("-", "_")
    model_name = model_name.lower()
    if model.version is not None and not omit_version:
        model_name += f":{model.version}"
    return model_name


def generate_ovms_model_address(ovms_address: str, model_name: str) -> str:
    """
    Generate a string representing a valid model address for the OpenVINO Model Server,
    from a given `ovms_address` and `model_name`.

    The format expected for model addresses by OVMS is:

        <address>:<port>/models/<model_name>[:<model_version>]

    Where `address` must not contain the protocol (http or https)

    :param ovms_address: IP address or URL pointing to the OVMS instance, including
        port specification
    :param model_name: Name of the model
    :return: String containing the model address
    """
    model_address = f"{ovms_address}/models/{model_name}"
    if model_address.startswith("https://"):
        model_address = model_address[8:]
    if model_address.startswith("http://"):
        model_address = model_address[7:]
    return model_address


def target_device_is_ovms(device: str) -> bool:
    """
    Return True if the target `device` specified is a URL or IP address, False otherwise

    :param device: Target device string to check
    :return: True if the string represents a URL or IP address, False otherwise
    """
    # Check if 'device' has been specified as a URL or IP address.
    server_pattern = re.compile(r"^((https?://)|(www.))(?:([a-zA-Z]+)|(\d+\.\d+\.\d+\.\d+)):\d{1,5}?$")
    return server_pattern.match(device) is not None


def rgb_to_hex(rgb: tuple[int, int, int]) -> str:
    """
    Convert an RGB color value to its corresponding hexadecimal representation.

    :param rgb: A tuple representing the RGB color value, where each element is an integer between 0 and 255.
    :return: The hexadecimal representation of the RGB color value.

    _Example:

        >>> rgb_to_hex((255, 0, 0))
        '#ff0000'
    """
    return f"#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}"


def assign_empty_label(prediction: Prediction, image: np.ndarray, empty_label: Label | None = None):
    """
    Modify the `prediction` in place to assign an empty label, if the prediction
    does not contain any objects.

    :param prediction: Prediction object to assign the empty label to, if needed
    :param image: The original image for which the prediction was generated
    :param empty_label: The Label object representing the empty label to assign
    """
    # Empty label is not generated by prediction postprocessing correctly, append
    # it here if there are no other predictions
    height, width = image.shape[0:2]
    if len(prediction.annotations) == 0 and empty_label is not None:
        prediction.append(
            Annotation(
                shape=Rectangle(x=0, y=0, width=width, height=height),
                labels=[ScoredLabel.from_label(empty_label, probability=1)],
            )
        )


def flow_control(intermediate_result: IntermediateInferenceResult, task: Task) -> tuple[list[ROI], list[np.ndarray]]:
    """
    Logic for the flow control tasks

    :param intermediate_result: IntermediateInferenceResult object, containing the
        predictions so far, as well as the rois.
    :param task: The task to perform the flow control operation for
    :return: Tuple containing:
        - rois -> A list of rois generated by the flow control task, to be passed
            down to the next task in the task chain
        - image_views -> A list of numpy arrays containing the image data in the
            corresponding roi
    """
    # CROP task
    if task.type == TaskType.CROP:
        rois = intermediate_result.filter_rois(label=None)
        image_views = intermediate_result.generate_views(rois)
    else:
        raise NotImplementedError(
            f"Unable to run inference for the pipeline in the deployed "
            f"project: Unsupported task type {task.type} found."
        )
    return rois, image_views


def pipeline_callback_factory(
    is_first_task: bool,
    is_final_task: bool,
    add_flow_control: bool,
    current_task: Task,
    deployment,
    final_callback: Callable[[Prediction, Any | None], None],
    next_trainable_task: Task | None = None,
    next_task: Task | None = None,
) -> Callable[[Prediction, Any], None]:
    """
    Generate callback functions for pipeline inference
    """
    if is_first_task:

        def callback(result: Prediction, runtime_data: tuple[np.ndarray, Any]):
            """
            Process inference results for the first task in a task chain, this
            initializes the intermediate result
            """
            image, runtime_user_data = runtime_data
            rois = [ROI.from_annotation(annotation) for annotation in result.annotations]

            if len(rois) == 0:
                # Empty prediction. Assign empty label and run final callback immediately
                assign_empty_label(result, image, deployment._empty_label)
                final_callback(image, result, runtime_user_data)
                return

            intermediate_result = IntermediateInferenceResult(image=image, prediction=result, rois=rois)

            if add_flow_control:
                rois, image_views = flow_control(intermediate_result, next_task)
                intermediate_result.rois = rois
            else:
                image_views = intermediate_result.generate_views()
            for roi, view in zip(rois, image_views):
                deployment._infer_task_async(
                    image=view,
                    explain=False,
                    task=next_trainable_task,
                    runtime_data=(image, roi, intermediate_result, runtime_user_data),
                )

    else:

        def callback(
            result: Prediction,
            runtime_data: tuple[np.ndarray, ROI, IntermediateInferenceResult, Any],
        ):
            """
            Process inference results for subsequent tasks
            """
            image, roi, intermediate_result, runtime_user_data = runtime_data

            if current_task.is_global:
                # Global tasks add their labels to the existing shape in the ROI
                intermediate_result.extend_annotations(result.annotations, roi=roi)
            else:
                # Local tasks create new shapes in the image coordinate system,
                # and generate ROI's corresponding to the new shapes
                for annotation in result.annotations:
                    intermediate_result.append_annotation(annotation, roi=roi)
                    new_roi = ROI.from_annotation(annotation)
                    intermediate_result.add_to_infer_queue(new_roi.to_absolute_coordinates(parent_roi=roi))
            intermediate_result.increment_infer_counter()
            if intermediate_result.all_rois_inferred():
                # In this case, all ROIS from the preceding task have been inferred.
                # The infer queue now becomes the new ROIs for the intermediate result
                new_rois = intermediate_result.get_infer_queue()
                intermediate_result.reset_infer_counter()
                intermediate_result.rois = new_rois

                if add_flow_control:
                    rois, image_views = flow_control(intermediate_result, next_task)
                    intermediate_result.rois = rois
                else:
                    image_views = intermediate_result.generate_views()

                if not is_final_task:
                    for roi, view in zip(rois, image_views):
                        deployment._infer_task_async(
                            image=view,
                            explain=False,
                            task=next_trainable_task,
                            runtime_data=(
                                image,
                                roi,
                                intermediate_result,
                                runtime_user_data,
                            ),
                        )
                else:
                    prediction = intermediate_result.prediction
                    final_callback(image, prediction, runtime_user_data)

    return callback



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from geti_sdk.http_session import GetiSession


class MultipleWorkspacesException(Exception):
    """Exception raised when multiple workspaces are available thus it is not possible to automatically select one."""

    def __init__(self, workspaces_list: list) -> None:
        ws_ids_and_names = [(ws["id"], ws["name"]) for ws in workspaces_list]
        error_message = (
            f"Multiple workspaces are available; please select one and provide its id through the 'workspace_id' "
            f"parameter when instantiating the client. Available workspaces (id, name): {ws_ids_and_names}."
        )
        super().__init__(error_message)


def get_workspace_id(rest_session: GetiSession) -> str:
    """
    Get the id of the workspace that is accessible with the provided session (cluster, organization, credentials, ...).

    In case of multiple workspaces, an error will be raised.

    :param rest_session: Session object which stores info about the connected cluster, organization and authentication
    :return: id of the workspace as a string
    :raises
    """
    workspaces = rest_session.get_rest_response(url="workspaces", method="GET")
    if isinstance(workspaces, list):
        workspace_list = workspaces
    elif isinstance(workspaces, dict):
        workspace_list = workspaces["workspaces"]
    else:
        raise ValueError(
            f"Unexpected response from cluster: {workspaces}. Expected to receive a "
            f"dictionary containing workspace data."
        )

    if len(workspace_list) > 1:
        raise MultipleWorkspacesException(workspace_list)

    return workspace_list[0]["id"]



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging

from geti_sdk.data_models import Project
from geti_sdk.data_models.configuration_models import Hyperparameters
from geti_sdk.data_models.configuration_models.training_configuration import (
    TrainingConfiguration,
)
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_converters.training_configuration_rest_converter import (
    TrainingConfigurationRESTConverter,
)


class TrainingConfigurationClient:
    """
    REST client for managing training configurations for machine learning models within a project.

    This client provides methods to retrieve and update training configurations, including
    full configurations and hyperparameters-only configurations for specific models.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        project_id = project.id
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project_id}/training_configuration"

    def get_configuration(self, model_manifest_id: str) -> TrainingConfiguration:
        """
        Retrieve the complete training configuration for a specific model manifest (AKA algorithm).

        Fetches the full training configuration including global parameters, hyperparameters,
        and all configuration sections (dataset preparation, training, evaluation) for the
        specified model manifest.

        :param model_manifest_id: Unique identifier of the model manifest to get configuration for
        :return: Complete TrainingConfiguration object containing all configuration parameters
        """
        url = f"{self.base_url}?model_manifest_id={model_manifest_id}"
        config_rest = self.session.get_rest_response(url=url, method="GET")
        config_rest["model_manifest_id"] = model_manifest_id
        return TrainingConfigurationRESTConverter.training_configuration_from_rest(config_rest)

    def get_hyperparameters_by_model_id(self, model_id: str, task_id: str | None) -> Hyperparameters:
        """
        Retrieve only the hyperparameters for a specific trained model.

        Fetches hyperparameters configuration for a specific model instance. This method
        returns only the hyperparameters portion of the configuration, excluding global
        parameters and other configuration metadata.

        :param model_id: Unique identifier of the trained model to get hyperparameters for
        :param task_id: Unique identifier of the task to get hyperparameters for
        :return: Hyperparameters object containing training, dataset preparation, and evaluation hyperparameters
        """
        url = f"{self.base_url}?model_id={model_id}"
        if task_id is not None:
            url += f"&task_id={task_id}"
        try:
            config_rest = self.session.get_rest_response(url=url, method="GET")
        except GetiRequestException as exc:
            if task_id is None:
                logging.error(
                    "Failed to get hyperparameters by model_id. Please note that for task-chain projects, "
                    "it's required to provide the task_id."
                )
            raise exc
        # when querying by model_id, the rest response contains only hyperparameters
        return TrainingConfigurationRESTConverter.hyperparameters_from_rest(config_rest)

    def set_configuration(self, configuration: TrainingConfiguration) -> None:
        """
        Update the training configuration for the project.

        Applies the provided training configuration to the project, updating all configuration
        parameters including global parameters and hyperparameters. The configuration will
        affect future training jobs in the project.

        :param configuration: TrainingConfiguration object containing the new configuration
                            parameters to apply to the project
        :return: None
        """
        config_rest = TrainingConfigurationRESTConverter.training_configuration_to_rest(configuration)
        self.session.get_rest_response(url=self.base_url, method="PATCH", data=config_rest)



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from copy import deepcopy

from geti_sdk.data_models.configuration_models import ProjectConfiguration, TaskConfig
from geti_sdk.rest_converters.configurable_parameters_rest_converter import (
    ConfigurableParametersRESTConverter,
)


class ProjectConfigurationRESTConverter(ConfigurableParametersRESTConverter):
    """
    Converter class for transforming ProjectConfiguration objects to/from REST API format.

    This class provides methods to convert between ProjectConfiguration domain objects and their
    corresponding REST representations, handling task configurations and auto-training settings.
    """

    @classmethod
    def task_config_to_rest(cls, task_config: TaskConfig) -> dict:
        """
        Convert a TaskConfig object to its REST API representation.

        Transforms a TaskConfig domain object into a dictionary structure expected by the
        REST API, including task_id and configurable parameters for training and auto_training.

        :param task_config: TaskConfig object containing training and auto-training configuration
        :return: Dictionary containing task_id, training parameters, and auto_training parameters
                in REST API format
        """
        return {
            "task_id": task_config.task_id,
            "training": cls.configurable_parameters_to_rest(task_config.training),
            "auto_training": cls.configurable_parameters_to_rest(task_config.auto_training),
        }

    @classmethod
    def project_configuration_to_rest(cls, project_configuration: ProjectConfiguration) -> dict:
        """
        Convert a ProjectConfiguration object to its REST API representation.

        Transforms a ProjectConfiguration domain object into a dictionary structure expected
        by the REST API, converting all task configurations to their REST format.

        :param project_configuration: ProjectConfiguration object containing all task configurations
        :return: Dictionary containing task_configs array with each task configuration converted
                to REST format
        """
        return {
            "task_configs": [
                cls.task_config_to_rest(task_config) for task_config in project_configuration.task_configs
            ],
        }

    @classmethod
    def project_configuration_from_rest(cls, rest_input: dict) -> ProjectConfiguration:
        """
        Create a ProjectConfiguration object from REST API input.

        Parses REST API dictionary and converts it into a validated ProjectConfiguration
        domain object, transforming each task configuration from REST format to domain format.

        :param rest_input: REST API dictionary containing task_configs array and other
                          project configuration parameters
        :return: Validated ProjectConfiguration object with all task configurations converted
                from REST format
        """
        rest_input = deepcopy(rest_input)
        task_configs = []
        for task_data in rest_input.pop("task_configs", {}):
            task_configs.append(cls.configurable_parameters_from_rest(task_data))

        return ProjectConfiguration.model_validate({"task_configs": task_configs} | rest_input)



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import copy
import logging
import warnings
from datetime import datetime
from typing import Any

import attrs
import cv2
import numpy as np

from geti_sdk.data_models import Dataset, Prediction, Project
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceAction
from geti_sdk.http_session import (
    GetiRequestException,
    GetiSession,
    ServerCredentialConfig,
    ServerTokenConfig,
)
from geti_sdk.rest_clients.dataset_client import DatasetClient
from geti_sdk.rest_clients.media_client.image_client import ImageClient
from geti_sdk.rest_clients.project_client.project_client import ProjectClient


class GetiDataCollection(PostInferenceAction):
    """
    Post inference action that will send an image to a specified `project` and `dataset`
    on the Intel® Geti™ server addressed by `session`.

    :param session: Geti session representing the connecting to the Intel® Geti™ server
    :param workspace_id: unique ID of the workspace in which the project to collect
        the data resides.
    :param project: Project or name of the project to whicht the image data should be
        transferred.
    :param dataset: Optional Dataset or name of the dataset in which to store the
        image data. If not specified, the default training dataset of the project is
        used
    :param log_level: Log level for the action. Options are 'info' or 'debug'
    """

    _override_from_dict_: bool = True

    def __init__(
        self,
        session: GetiSession,
        workspace_id: str,
        project: str | Project,
        dataset: str | Dataset | None = None,
        log_level: str = "debug",
    ):
        super().__init__(log_level=log_level)
        project_client = ProjectClient(session=session, workspace_id=workspace_id)
        if isinstance(project, str):
            project_name = project
            project = project_client.get_project_by_name(project_name=project_name)
            if project is None:
                raise ValueError(
                    f"Project `{project_name}` does not exist on the Intel® Geti™ "
                    f"server, unable to initialize the Intel® Geti™ data collection "
                    f"action"
                )
        dataset_client = DatasetClient(session=session, workspace_id=workspace_id, project=project)
        if dataset is None:
            datasets = dataset_client.get_all_datasets()
            dataset = [ds for ds in datasets if ds.use_for_training][0]
        elif isinstance(dataset, str):
            dataset_name = dataset

            try:
                dataset = dataset_client.get_dataset_by_name(dataset_name)
            except ValueError:
                dataset = dataset_client.create_dataset(dataset_name)
                self.log_function(f"Dataset `{dataset_name}` was created in project `{project.name}`")
        self.image_client = ImageClient(session=session, workspace_id=workspace_id, project=project)
        self.dataset = dataset
        self._repr_info_ = (
            f"target_server=`{session.config.host}`, target_project={project.name}, target_dataset={dataset.name}"
        )

        # Serialize input arguments
        self._constructor_arguments_["project"] = project.name
        self._constructor_arguments_["dataset"] = dataset.name
        self._constructor_arguments_["session"] = attrs.asdict(session.config)

    def __call__(
        self,
        image: np.ndarray,
        prediction: Prediction,
        score: float | None = None,
        name: str | None = None,
        timestamp: datetime | None = None,
    ):
        """
        Execute the action, upload the given `image` to the Intel® Geti™ server.

        The parameters `prediction`, `score`, `name` and `timestamp` are not used in
        this specific action.

        :param image: Numpy array representing an image
        :param prediction: Prediction object which was generated for the image
        :param score: Optional score computed from a post inference trigger
        :param name: String containing the name of the image
        :param timestamp: Datetime object containing the timestamp belonging to the
            image
        """
        image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
        # upload_image uses cv2 to encode the numpy array as image, so it expects an
        # image in BGR format. However, `Deployment.infer` requires RGB format, so
        # we have to convert
        try:
            self.image_client.upload_image(image=image_bgr, dataset=self.dataset)
        except GetiRequestException as e:
            logging.exception(e)
        self.log_function(f"GetiDataCollection inference action uploaded image to dataset `{self.dataset.name}`")

    @classmethod
    def from_dict(cls, input_dict: dict[str, Any]) -> "GetiDataCollection":
        """
        Construct a GetiDataCollection post inference action object from an input
        dictionary `input_dict`

        :param input_dict: Dictionary representation of the GetiDataCollection
        :return: Instantiated GetiDataCollection, according to the input dictionary
        """
        input_copy = copy.deepcopy(input_dict)
        session_dict = input_copy["session"]
        if "username" in session_dict:
            server_config_class = ServerCredentialConfig
        elif "token" in session_dict:
            server_config_class = ServerTokenConfig
        else:
            raise ValueError(f"Invalid `GetiSession` parameters encountered: {session_dict}")
        session = GetiSession(server_config_class(**session_dict))
        input_copy.update({"session": session})
        return cls(**input_copy)

    def to_dict(self) -> dict[str, Any]:
        """
        Return a dictionary representation of the PostInferenceObject

        :return: Dictionary representing the class name and its constructor parameters
        """
        warnings.warn(
            "GetiDataCollection post inference action contains sensitive information "
            "used for authentication on the Intel® Geti™ platform. Be careful when "
            "saving this information to disk or sharing with others!"
        )
        return super().to_dict()



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Deprecation utilities for Geti SDK.

This module provides utilities for marking functions and classes as deprecated.
The utilities help maintain backward compatibility while encouraging users to
migrate to newer APIs.

Example:
    >>> from geti_sdk.utils.deprecation import deprecated

    >>> # Deprecation without any information
    >>> @deprecate
    ... def old_function():
    ...     pass

    >>> # Deprecation with replacement function
    >>> @deprecate(use="new_function")
    ... def old_function():
    ...     pass

    >>> # Deprecation with version info and replacement class
    >>> @deprecate(since="2.1.0", remove="2.5.0", use="NewClass")
    ... class OldClass:
    ...     pass

    >>> # Deprecation with only removal version
    >>> @deprecate(remove="2.5.0")
    ... def another_function():
    ...     pass

    >>> # Deprecation with only deprecation version
    >>> @deprecate(since="2.1.0")
    ... def yet_another_function():
    ...     pass
"""

import inspect
import warnings
from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar, overload

_T = TypeVar("_T")


@overload
def deprecate(__obj: _T) -> _T: ...


@overload
def deprecate(
    *,
    since: str | None = None,
    remove: str | None = None,
    use: str | None = None,
    reason: str | None = None,
    warning_category: type[Warning] = FutureWarning,
) -> Callable[[_T], _T]: ...


def deprecate(
    __obj: Any = None,
    *,
    since: str | None = None,
    remove: str | None = None,
    use: str | None = None,
    reason: str | None = None,
    warning_category: type[Warning] = FutureWarning,
) -> Any:
    """
    Mark a function or class as deprecated.

    This decorator will cause a warning to be emitted when the function is called or class is instantiated.

    Example:
        >>> # Basic deprecation without any information
        >>> @deprecate
        ... def old_function():
        ...     pass

        >>> # Deprecation with version info and replacement class
        >>> @deprecate(since="2.1.0", remove="2.5.0", use="NewClass")
        ... class OldClass:
        ...     pass

        >>> # Deprecation with only removal version
        >>> @deprecate(remove="2.5.0")
        ... def another_function():
        ...     pass

        >>> # Deprecation with only deprecation version
        >>> @deprecate(since="2.1.0")
        ... def yet_another_function():
        ...     pass

        >>> # Deprecation with reason and custom warning category
        >>> @deprecate(since="2.1.0", reason="Performance improvements available", warning_category=FutureWarning)
        ... def old_slow_function():
        ...     pass

    :param __obj: The function or class to be deprecated.
            If provided, the decorator is used without arguments.
            If None, the decorator is used with arguments.
    :param since: Version when the function/class was deprecated (e.g. "2.1.0").
            If not provided, no deprecation version will be shown in the warning message.
    :param remove: Version when the function/class will be removed.
            If not provided, no removal version will be shown in the warning message.
    :param use: Name of the replacement function/class.
            If not provided, no replacement suggestion will be shown in the warning message.
    :param reason: Additional reason for the deprecation.
            If not provided, no reason will be shown in the warning message.
    :param warning_category: Type of warning to emit (default: FutureWarning).
            Can be DeprecationWarning, FutureWarning, PendingDeprecationWarning, etc.
    :return: Decorated function/class that emits a deprecation warning when used.
    """

    def _deprecate_impl(obj: Any) -> Any:  # noqa: ANN401
        if isinstance(obj, type):
            # Handle class deprecation
            original_init = inspect.getattr_static(obj, "__init__")

            @wraps(original_init)
            def new_init(self: Any, *args: Any, **kwargs: Any) -> None:  # noqa: ANN401
                msg = f"{obj.__name__} is deprecated"
                if since:
                    msg += f" since v{since}"
                if remove:
                    msg += f" and will be removed in v{remove}"
                if use:
                    msg += f". Use {use} instead"
                if reason:
                    msg += f". {reason}"
                msg += "."
                warnings.warn(msg, warning_category, stacklevel=2)
                original_init(self, *args, **kwargs)

            setattr(obj, "__init__", new_init)  # noqa: B010
            return obj

        # Handle function deprecation
        @wraps(obj)
        def wrapper(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
            msg = f"{obj.__name__} is deprecated"
            if since:
                msg += f" since v{since}"
            if remove:
                msg += f" and will be removed in v{remove}"
            if use:
                msg += f". Use {use} instead"
            if reason:
                msg += f". {reason}"
            msg += "."
            warnings.warn(msg, warning_category, stacklevel=2)
            return obj(*args, **kwargs)

        return wrapper

    if __obj is None:
        return _deprecate_impl
    return _deprecate_impl(__obj)



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from pydantic import BaseModel, Field, model_validator


class TrainConstraints(BaseModel):
    """Constraints applied for model training."""

    min_images_per_label: int = Field(
        ge=0,
        title="Minimum number of images per label",
        description="Minimum number of images that must be present for each label to train",
    )


class AutoTrainingParameters(BaseModel):
    """Configuration for auto-training feature."""

    enable: bool = Field(
        title="Enable auto training",
        description="Whether automatic training is enabled for this task",
    )
    enable_dynamic_required_annotations: bool = Field(
        title="Enable dynamic required annotations",
        description="Whether to dynamically adjust the number of required annotations",
    )
    min_images_per_label: int = Field(
        ge=0,
        title="Minimum images per label",
        description="Minimum number of images needed for each label to trigger auto-training",
    )


class TrainingParameters(BaseModel):
    """Parameters that control the training process."""

    constraints: TrainConstraints = Field(
        title="Training constraints",
        description="Constraints that must be satisfied for training to proceed",
    )


class TaskConfig(BaseModel):
    """Configuration for a specific task within a project."""

    task_id: str = Field(title="Task ID", description="Unique identifier for the task")
    training: TrainingParameters = Field(
        title="Training parameters",
        description="Parameters controlling the training process",
    )
    auto_training: AutoTrainingParameters = Field(
        title="Auto-training parameters",
        description="Parameters controlling auto-training",
    )

    @model_validator(mode="after")
    def task_id_not_empty(self) -> "TaskConfig":
        if not self.task_id:
            raise ValueError("Task ID must be provided as part of the task configuration and cannot be empty.")
        return self


class ProjectConfiguration(BaseModel):
    """
    Configurable parameters for a project.

    Each project has exactly one configuration entity. The ID of this entity
    matches the project ID, as there is a one-to-one relationship between
    projects and their configurations.
    """

    task_configs: list[TaskConfig] = Field(
        title="Task configurations",
        description="List of configurations for all tasks in this project",
    )

    def get_task_config(self, task_id: str) -> TaskConfig:
        """
        Retrieves the configuration for a specific task by its ID.

        :param task_id: The ID of the task to retrieve the configuration for.
        :return: The TaskConfig for the specified task
        :raises ValueError: If the task ID is not found in the project configuration.
        """
        for task_conf in self.task_configs:
            if task_conf.task_id == task_id:
                return task_conf
        raise ValueError(f"Task configuration with ID {task_id} not found.")



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum
from pprint import pformat
from typing import Any

import attr
from pydantic import BaseModel, Field

from geti_sdk.data_models.enums import TaskType
from geti_sdk.data_models.utils import str_to_optional_enum_converter

from .utils import attr_value_serializer, remove_null_fields


@attr.define
class LegacyAlgorithm:
    """
    Representation of a supported algorithm on the Intel® Geti™ platform.
    """

    model_size: str
    model_template_id: str
    gigaflops: float
    name: str | None = None
    summary: str | None = None
    task_type: str | None = attr.field(default=None, converter=str_to_optional_enum_converter(TaskType))
    supports_auto_hpo: bool | None = None  # Deprecated in Geti v2.7
    default_algorithm: bool | None = None  # Added in Geti v1.16
    performance_category: str | None = None  # Added in Geti v1.9
    lifecycle_stage: str | None = None  # Added in Geti v1.9

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the Algorithm to a dictionary representation.

        :return: Dictionary holding the algorithm data
        """
        return attr.asdict(self, value_serializer=attr_value_serializer)

    @property
    def overview(self) -> str:
        """
        Return a string that shows an overview of the Algorithm properties.

        :return: String holding an overview of the algorithm
        """
        overview_dict = self.to_dict()
        remove_null_fields(overview_dict)
        return pformat(overview_dict)


class GPUMaker(str, Enum):
    """GPU maker names."""

    NVIDIA = "nvidia"
    INTEL = "intel"

    def __str__(self) -> str:
        """Returns the name of the GPU maker."""
        return str(self.name)


class AlgorithmDeprecationStatus(str, Enum):
    """Status of a model architecture with respect to the deprecation process."""

    ACTIVE = "active"  # Model architecture is fully supported, models can be trained
    DEPRECATED = "deprecated"  # Model architecture is deprecated, can still view and train but it's discouraged
    OBSOLETE = "obsolete"  # Model architecture is no longer supported, models can be still viewed but not trained

    def __str__(self) -> str:
        """Returns the name of the model status."""
        return str(self.name)


class PerformanceRatings(BaseModel):
    """Ratings for different performance aspects of a model."""

    accuracy: int = Field(
        ge=1,
        le=3,
        title="Accuracy rating",
        description="Rating of the model accuracy. "
        "The value should be interpreted relatively to the other available models, "
        "and it ranges from 1 (below average) to 3 (above average).",
    )
    training_time: int = Field(
        ge=1,
        le=3,
        title="Training time rating",
        description="Rating of the model training time. "
        "The value should be interpreted relatively to the other available models, "
        "and it ranges from 1 (below average/slower) to 3 (above average/faster).",
    )
    inference_speed: int = Field(
        ge=1,
        le=3,
        title="Inference speed rating",
        description="Rating of the model inference speed. "
        "The value should be interpreted relatively to the other available models, "
        "and it ranges from 1 (below average/slower) to 3 (above average/faster).",
    )


class ModelStats(BaseModel):
    """Information about a machine learning model."""

    gigaflops: float = Field(
        ge=0,
        title="Gigaflops",
        description="Billions of floating-point operations per second required by the model",
    )
    trainable_parameters: float = Field(
        ge=0.0,
        title="Trainable parameters (millions)",
        description="Number of trainable parameters in the model, expressed in millions",
    )
    performance_ratings: PerformanceRatings = Field(
        title="Performance ratings",
        description="Standardized ratings for model performance metrics",
    )


class Capabilities(BaseModel):
    """Model capabilities configuration."""

    xai: bool = Field(
        title="Explainable AI Support",
        description="Whether the model supports explainable AI features",
    )
    tiling: bool = Field(
        title="Tiling Support",
        description="Whether the model supports image tiling for processing large images",
    )


class Algorithm(BaseModel):
    model_manifest_id: str
    task: TaskType
    name: str
    description: str
    stats: ModelStats
    support_status: AlgorithmDeprecationStatus
    supported_gpus: dict[GPUMaker, bool]
    capabilities: Capabilities
    is_default_model: bool
    performance_category: str

    @staticmethod
    def from_legacy_algorithm(legacy_algorithm: LegacyAlgorithm) -> "Algorithm":
        """
        Convert a LegacyAlgorithm to an Algorithm.

        :param legacy_algorithm: LegacyAlgorithm to convert
        :return: Algorithm representation of the legacy algorithm
        """
        return Algorithm(
            model_manifest_id=legacy_algorithm.model_template_id,
            task=legacy_algorithm.task,
            name=legacy_algorithm.name or "Unnamed Algorithm",
            description=legacy_algorithm.summary or "",
            stats=ModelStats(
                gigaflops=legacy_algorithm.gigaflops,
                trainable_parameters=0.0,  # Trainable parameters not available in legacy algorithm
                performance_ratings=PerformanceRatings(
                    accuracy=1,  # Default value, as legacy does not provide ratings
                    training_time=1,
                    inference_speed=1,
                ),
            ),
            support_status=AlgorithmDeprecationStatus[legacy_algorithm.lifecycle_stage],
            supported_gpus={
                GPUMaker.NVIDIA: True,  # Default assumption, as legacy does not specify
                GPUMaker.INTEL: False,  # Default assumption, as legacy does not specify
            },
            capabilities=Capabilities(xai=False, tiling=False),  # Default capabilities
            is_default_model=legacy_algorithm.default_algorithm,
            performance_category=legacy_algorithm.performance_category,
        )



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from geti_sdk.data_models import Project
from geti_sdk.data_models.configuration_models.project_configuration import (
    ProjectConfiguration,
)
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_converters.project_configuration_rest_converter import (
    ProjectConfigurationRESTConverter,
)


class ProjectConfigurationClient:
    """
    REST client for managing project-level configuration settings.

    This client provides methods to retrieve and update project configuration,
    including task configurations and auto-training settings for all tasks
    within a project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        project_id = project.id
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project_id}/project_configuration"

    def get_configuration(self) -> ProjectConfiguration:
        """
        Retrieve the complete project configuration.

        Fetches the current project configuration including all task configurations,
        auto-training settings, and other project-level configuration parameters.

        :return: ProjectConfiguration object containing all project-level settings
        """
        config_rest = self.session.get_rest_response(url=self.base_url, method="GET")
        return ProjectConfigurationRESTConverter.project_configuration_from_rest(config_rest)

    def set_project_auto_train(self, auto_train: bool = False) -> None:
        """
        Enable or disable auto-training for all tasks in the project.

        Updates the auto_training.enable setting for every task configuration
        in the project to the specified value. This is a convenience method
        to bulk update auto-training settings across all tasks.

        :param auto_train: True to enable auto-training for all tasks, False to disable
        :return: None
        """
        project_configuration = self.get_configuration()
        for task_config in project_configuration.task_configs:
            task_config.auto_training.enable = auto_train
        self.set_configuration(project_configuration)

    def set_configuration(self, configuration: ProjectConfiguration) -> None:
        """
        Update the project configuration with new settings.

        Applies the provided project configuration to the project, updating all
        configuration parameters including task configurations and auto-training
        settings. Changes will affect the behavior of all tasks in the project.

        :param configuration: ProjectConfiguration object containing the new configuration
                            settings to apply to the project
        :return: None
        """
        config_rest = ProjectConfigurationRESTConverter.project_configuration_to_rest(configuration)
        self.session.get_rest_response(url=self.base_url, method="PATCH", data=config_rest)



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from pydantic import BaseModel, Field, model_validator

from .augmentation import AugmentationParameters


class DatasetPreparationParameters(BaseModel):
    """Parameters for dataset preparation before training."""

    augmentation: AugmentationParameters = Field(
        default_factory=AugmentationParameters,
        title="Data augmentation",
        description="Configuration for data augmentation techniques applied to the dataset",
    )


class EarlyStopping(BaseModel):
    enable: bool = Field(
        default=False,
        title="Enable early stopping",
        description="Whether to stop training early when performance stops improving",
    )
    patience: int = Field(
        gt=0,
        default=1,
        title="Patience",
        description="Number of epochs with no improvement after which training will be stopped",
    )


class TrainingHyperParameters(BaseModel):
    """Hyperparameters for model training process."""

    max_epochs: int = Field(
        gt=0,
        default=1000,
        title="Maximum epochs",
        description="Maximum number of training epochs to run",
    )
    early_stopping: EarlyStopping = Field(
        default_factory=EarlyStopping,
        title="Early stopping",
        description="Configuration for early stopping mechanism",
    )
    learning_rate: float = Field(
        gt=0,
        lt=1,
        default=0.001,
        title="Learning rate",
        description="Base learning rate for the optimizer",
    )
    input_size_width: int | None = Field(
        default=None,
        gt=0,
        title="Input size width",
        description=(
            "Width dimension in pixels for model input images. "
            "Determines the horizontal resolution at which images are processed."
        ),
        json_schema_extra={},
    )
    input_size_height: int | None = Field(
        default=None,
        gt=0,
        title="Input size height",
        description=(
            "Height dimension in pixels for model input images. "
            "Determines the vertical resolution at which images are processed."
        ),
        json_schema_extra={},
    )
    allowed_values_input_size: list[int] | None = Field(
        default=None,
        title="Supported input size dimensions",
        description=(
            "List of supported values for input width and height. "
            "When specified, both width and height must be chosen from these values."
        ),
        json_schema_extra={"validation_only": True},
    )

    @model_validator(mode="after")
    def validate_input_size(self) -> "TrainingHyperParameters":
        w, h = self.input_size_width, self.input_size_height

        # Skip validation if neither width nor height are set
        if w is None and h is None:
            return self

        # For non-partial models, both width and height must be set
        class_name = type(self).__name__
        if "partial" not in class_name.lower() and (w is None or h is None):
            raise ValueError("Both input_size_width and input_size_height must be specified")

        # Validate against allowed input sizes if available
        if allowed_values := self.allowed_values_input_size:
            if w and w not in allowed_values:
                raise ValueError(
                    f"Input size width '{w}' is not in the list of supported input sizes: {allowed_values}"
                )
            if h and h not in allowed_values:
                raise ValueError(
                    f"Input size height '{h}' is not in the list of supported input sizes: {allowed_values}"
                )

        # Update the model's json_schema to include allowed values for input_size
        # Note: existing json_schema_extra cannot be overwritten, otherwise it will absent from model_json_schema()
        self.model_fields["input_size_width"].json_schema_extra.update(  # type: ignore[union-attr]
            {
                "allowed_values": self.allowed_values_input_size,  # type:ignore[dict-item]
                "default_value": w,
            }
        )
        self.model_fields["input_size_height"].json_schema_extra.update(  # type: ignore[union-attr]
            {
                "allowed_values": self.allowed_values_input_size,  # type:ignore[dict-item]
                "default_value": h,
            }
        )
        return self


class EvaluationParameters(BaseModel):
    """Parameters for model evaluation."""

    metric: str | None = Field(
        default=None,
        title="Evaluation metric",
        description="Metric used to evaluate model performance",
    )


class Hyperparameters(BaseModel):
    """Complete set of configurable parameters for model training and evaluation."""

    dataset_preparation: DatasetPreparationParameters = Field(
        default_factory=DatasetPreparationParameters,
        title="Dataset preparation",
        description="Parameters for preparing the dataset before training",
    )
    training: TrainingHyperParameters | None = Field(
        default=None,
        title="Training hyperparameters",
        description="Hyperparameters for the model training process",
    )
    evaluation: EvaluationParameters = Field(
        default_factory=EvaluationParameters,
        title="Evaluation parameters",
        description="Parameters for evaluating the trained model",
    )



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Any

from pydantic import BaseModel

PYDANTIC_BASE_TYPES = {
    "integer",
    "number",
    "boolean",
    "string",
}
PYDANTIC_ANY_OF = "anyOf"


class ConfigurableParametersRESTConverter:
    """
    Base class for converting configurable parameters to REST views.

    This class provides methods to transform Pydantic models and their fields
    into REST-compatible dictionary representations.
    """

    @classmethod
    def configurable_parameters_to_rest(
        cls, configurable_parameters: BaseModel
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """
        Convert a Pydantic model of configurable parameters to its REST representation.

        This method processes a Pydantic model containing configuration parameters and transforms it
        into a REST view. It handles both simple fields and nested models:

        - Simple fields (int, float, str, bool) are converted to a list of dictionaries with metadata
            including key, name, description, value, type, and constraints
        - Nested Pydantic models are processed recursively and maintained as nested structures

        The return format depends on the content:
        - If only simple parameters exist: returns a list of parameter dictionaries
        - If only nested models exist: returns a dictionary mapping nested model names to their contents
        - If both exist: returns a list containing parameter dictionaries and nested model dictionary

        :param configurable_parameters: Pydantic model containing configurable parameters
        :return: REST representation as either a dictionary of nested models,
            a list of parameter dictionaries, or a combined list of both
        """
        nested_params: dict[str, Any] = {}
        list_params: list[dict[str, Any]] = []

        for field_name in configurable_parameters.model_fields:
            if field_name.startswith("allowed_values_"):
                # Skip fields that are not part of the main configurable parameters
                continue

            field = getattr(configurable_parameters, field_name)
            if isinstance(field, BaseModel):
                # If the field is a nested Pydantic model, process it recursively
                nested_params[field_name] = cls.configurable_parameters_to_rest(configurable_parameters=field)
            else:
                # If the field is a simple type, convert directly to REST view
                list_params.append(
                    {
                        "key": field_name,
                        "value": field,
                    }
                )

        # Return combined or individual results based on content
        if nested_params and list_params:
            return [*list_params, nested_params]
        return list_params or nested_params

    @classmethod
    def configurable_parameters_from_rest(
        cls, configurable_parameters_rest: dict[str, Any] | list[dict[str, Any]]
    ) -> dict[str, Any]:
        """
        Convert a REST representation back to a dictionary that can be used to create/update a Pydantic model.

        This method performs the reverse operation of configurable_parameters_to_rest:
        - For a list of parameter dictionaries, it extracts the key-value pairs
        - For a dictionary of nested models, it processes each nested model recursively
        - For a mixed list containing both, it handles both types

        :param configurable_parameters_rest: REST representation as a dictionary or list
        :return: Dictionary representation suitable for Pydantic model instantiation
        """
        # If the input is a list (of parameters or mixed)
        if isinstance(configurable_parameters_rest, list):
            result = {}

            for item in configurable_parameters_rest:
                # If this is a parameter entry (has a "key" field)
                if isinstance(item, dict) and "key" in item:
                    key = item["key"]
                    value = item["value"]
                    result[key] = value
                # If it's a dictionary without a "key" field, it must contain nested models
                elif isinstance(item, dict):
                    # Process each nested model recursively and merge with result
                    nested_result = cls.configurable_parameters_from_rest(item)
                    result.update(nested_result)

            return result

        # If the input is a dictionary (of nested models or other fields)
        if isinstance(configurable_parameters_rest, dict):
            result = {}

            for key, value in configurable_parameters_rest.items():
                # If the value is a complex structure, process it recursively
                if isinstance(value, dict | list):  # Replace `dict | list` with `(dict, list)`
                    result[key] = cls.configurable_parameters_from_rest(value)
                else:
                    # Simple value, keep as is
                    result[key] = value

            return result

        # If it's neither a list nor a dictionary, return it as is
        return configurable_parameters_rest



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from collections import defaultdict
from copy import deepcopy
from typing import Any

from geti_sdk.data_models.configuration_models import (
    DatasetPreparationParameters as HyperparametersDatasetPreparationParameters,
)
from geti_sdk.data_models.configuration_models import (
    EvaluationParameters,
    GlobalDatasetPreparationParameters,
    GlobalParameters,
    Hyperparameters,
    TrainingConfiguration,
    TrainingHyperParameters,
)
from geti_sdk.rest_converters.configurable_parameters_rest_converter import (
    ConfigurableParametersRESTConverter,
)

DATASET_PREPARATION = "dataset_preparation"
TRAINING = "training"
EVALUATION = "evaluation"


class TrainingConfigurationRESTConverter(ConfigurableParametersRESTConverter):
    """
    Converter class for transforming TrainingConfiguration objects to/from REST API format.

    This class handles the conversion between TrainingConfiguration domain objects and their
    REST API representations, managing the separation and combination of global parameters
    and hyperparameters across different configuration sections (dataset preparation,
    training, and evaluation).
    """

    @classmethod
    def _dataset_preparation_to_rest(
        cls,
        global_parameters: GlobalParameters | None,
        hyperparameters: Hyperparameters | None,
    ) -> dict[str, Any]:
        """
        Convert dataset preparation parameters from both global and hyperparameters to REST format.

        Merges dataset preparation parameters from global parameters and hyperparameters
        into a single dictionary for REST API consumption. Global parameters take precedence
        over hyperparameters when both contain the same keys.

        :param global_parameters: Global parameters object containing dataset preparation settings,
                                 None if not available
        :param hyperparameters: Hyperparameters object containing dataset preparation settings,
                               None if not available
        :return: Dictionary containing merged dataset preparation parameters for REST API
        :raises ValueError: If the parameter conversion doesn't result in dictionaries
        """
        # Return a combined view of global and hyperparameters for dataset preparation
        global_parameters_rest = (
            cls.configurable_parameters_to_rest(
                configurable_parameters=global_parameters.dataset_preparation,
            )
            if global_parameters and global_parameters.dataset_preparation
            else {}
        )
        hyperparameters_rest = (
            cls.configurable_parameters_to_rest(
                configurable_parameters=hyperparameters.dataset_preparation,
            )
            if hyperparameters and hyperparameters.dataset_preparation
            else {}
        )
        if not isinstance(global_parameters_rest, dict) or not isinstance(hyperparameters_rest, dict):
            raise ValueError("Expected dictionary for global and hyperparameters REST views")
        return global_parameters_rest | hyperparameters_rest

    @classmethod
    def training_configuration_to_rest(cls, training_configuration: TrainingConfiguration) -> dict[str, Any]:
        """
        Convert a TrainingConfiguration object to its REST API representation.

        Transforms the training configuration into a dictionary structure expected by the
        REST API, including task_id, model_manifest_id, and parameter sections for
        dataset_preparation, training, and evaluation.

        :param training_configuration: TrainingConfiguration or PartialTrainingConfiguration
                                     object to convert to REST format
        :return: Dictionary containing the complete REST representation with task_id,
                model_manifest_id, and parameter sections (dataset_preparation, training, evaluation)
        """
        training_params_rest = (
            cls.configurable_parameters_to_rest(
                configurable_parameters=training_configuration.hyperparameters.training,
            )
            if training_configuration.hyperparameters and training_configuration.hyperparameters.training
            else []
        )

        return {
            "task_id": training_configuration.task_id,
            "model_manifest_id": training_configuration.model_manifest_id,
            DATASET_PREPARATION: cls._dataset_preparation_to_rest(
                global_parameters=training_configuration.global_parameters,
                hyperparameters=training_configuration.hyperparameters,
            ),
            TRAINING: training_params_rest,
            EVALUATION: [],  # Evaluation parameters are not yet available
        }

    @classmethod
    def configuration_dict_from_rest(cls, rest_input: dict[str, Any]) -> dict[str, Any]:
        """
        Parse REST input and reorganize it into a structure suitable for Pydantic model creation.

        Separates the flat REST parameter structure into global_parameters and hyperparameters
        sections based on the model field definitions. This method handles the reverse operation
        of training_configuration_to_rest by organizing parameters by their proper domains.

        :param rest_input: REST API dictionary containing parameters organized by sections
                          (dataset_preparation, training, evaluation) plus task_id and model_manifest_id
        :return: Dictionary with 'global_parameters' and 'hyperparameters' keys properly organized
                for TrainingConfiguration model validation, plus any additional fields from rest_input
        """
        rest_input = deepcopy(rest_input)
        dataset_preparation = cls.configurable_parameters_from_rest(rest_input.pop(DATASET_PREPARATION, {}))
        training = cls.configurable_parameters_from_rest(rest_input.pop(TRAINING, {}))
        evaluation = cls.configurable_parameters_from_rest(rest_input.pop(EVALUATION, {}))

        global_parameters: dict = defaultdict(dict)
        hyperparameters: dict = defaultdict(dict)

        for field, _ in GlobalDatasetPreparationParameters.model_fields.items():
            global_parameters[DATASET_PREPARATION][field] = dataset_preparation.pop(field, None)

        for (
            field,
            _,
        ) in HyperparametersDatasetPreparationParameters.model_fields.items():
            hyperparameters[DATASET_PREPARATION][field] = dataset_preparation.pop(field, None)

        for field, _ in TrainingHyperParameters.model_fields.items():
            hyperparameters[TRAINING][field] = training.pop(field, None)

        for field, _ in EvaluationParameters.model_fields.items():
            hyperparameters[EVALUATION][field] = evaluation.pop(field, None)

        # add remaining parameters for validation (extra parameters should not be present)
        global_parameters[DATASET_PREPARATION].update(dataset_preparation)
        hyperparameters[TRAINING].update(training)
        hyperparameters[EVALUATION].update(evaluation)

        # Convert defaultdict to regular dicts for the model validation
        global_parameters = dict(global_parameters)
        hyperparameters = dict(hyperparameters)
        global_parameters.pop("default_factory", None)
        hyperparameters.pop("default_factory", None)

        return {
            "global_parameters": global_parameters,
            "hyperparameters": hyperparameters,
        } | rest_input

    @classmethod
    def training_configuration_from_rest(cls, rest_input: dict[str, Any]) -> TrainingConfiguration:
        """
        Create a TrainingConfiguration object from REST API input.

        Combines the parsing and validation steps to convert a REST API dictionary
        directly into a validated TrainingConfiguration domain object.

        :param rest_input: REST API dictionary containing all configuration data including
                          task_id, model_manifest_id, and parameter sections
        :return: Validated TrainingConfiguration object created from the REST input
        """
        dict_model = cls.configuration_dict_from_rest(rest_input)
        return TrainingConfiguration.model_validate(dict_model)

    @classmethod
    def hyperparameters_from_rest(cls, rest_input: dict[str, Any]) -> Hyperparameters:
        """
        Extract and create a Hyperparameters object from REST API input.

        Parses the REST input to extract only the hyperparameters portion, ignoring
        global parameters and other configuration data. Useful when only hyperparameters
        are needed from a full configuration.

        :param rest_input: REST API dictionary containing configuration parameters
        :return: Validated Hyperparameters object containing training, dataset preparation,
                and evaluation hyperparameters extracted from the REST input
        """
        dict_model = cls.configuration_dict_from_rest(rest_input)["hyperparameters"]
        return Hyperparameters.model_validate(dict_model)



# Copyright (C) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


from pydantic import BaseModel, ConfigDict, Field

from .hyperparameters.hyperparameters import Hyperparameters


class SubsetSplit(BaseModel):
    """
    Parameters for splitting a dataset into training, validation, and test subsets.
    The sum of training, validation, and test percentages must equal 100.
    """

    training: int = Field(
        ge=1,
        le=100,
        title="Training percentage",
        description="Percentage of data to use for training",
    )
    validation: int = Field(
        ge=1,
        le=100,
        title="Validation percentage",
        description="Percentage of data to use for validation",
    )
    test: int = Field(
        ge=1,
        le=100,
        title="Test percentage",
        description="Percentage of data to use for testing",
    )
    auto_selection: bool = Field(
        title="Auto selection",
        description="Whether to automatically select data for each subset",
    )
    remixing: bool = Field(
        default=False,
        title="Remixing",
        description="Whether to remix data between subsets",
    )


class MinAnnotationPixels(BaseModel):
    """Parameters for minimum annotation pixels."""

    enable: bool = Field(
        default=False,
        title="Enable minimum annotation pixels filtering",
        description="Whether to apply minimum annotation pixels filtering",
    )
    min_annotation_pixels: int = Field(
        gt=0,
        le=200000000,  # reasonable upper limit for pixel count to 200MP
        default=1,
        title="Minimum annotation pixels",
        description="Minimum number of pixels in an annotation",
    )


class MaxAnnotationPixels(BaseModel):
    """Parameters for maximum annotation pixels."""

    enable: bool = Field(
        default=False,
        title="Enable maximum annotation pixels filtering",
        description="Whether to apply maximum annotation pixels filtering",
    )
    max_annotation_pixels: int = Field(
        gt=0,
        default=10000,
        title="Maximum annotation pixels",
        description="Maximum number of pixels in an annotation",
    )


class MinAnnotationObjects(BaseModel):
    """Parameters for maximum annotation objects."""

    enable: bool = Field(
        default=False,
        title="Enable minimum annotation objects filtering",
        description="Whether to apply minimum annotation objects filtering",
    )
    min_annotation_objects: int = Field(
        gt=0,
        default=1,
        title="Minimum annotation objects",
        description="Minimum number of objects in an annotation",
    )


class MaxAnnotationObjects(BaseModel):
    """Parameters for maximum annotation objects."""

    enable: bool = Field(
        default=False,
        title="Enable maximum annotation objects filtering",
        description="Whether to apply maximum annotation objects filtering",
    )
    max_annotation_objects: int = Field(
        gt=0,
        default=10000,
        title="Maximum annotation objects",
        description="Maximum number of objects in an annotation",
    )


class Filtering(BaseModel):
    """Parameters for filtering annotations in the dataset."""

    min_annotation_pixels: MinAnnotationPixels = Field(
        title="Minimum annotation pixels",
        description="Minimum number of pixels in an annotation",
    )
    max_annotation_pixels: MaxAnnotationPixels = Field(
        title="Maximum annotation pixels",
        description="Maximum number of pixels in an annotation",
    )
    min_annotation_objects: MinAnnotationObjects = Field(
        title="Minimum annotation objects",
        description="Minimum number of objects in an annotation",
    )
    max_annotation_objects: MaxAnnotationObjects = Field(
        title="Maximum annotation objects",
        description="Maximum number of objects in an annotation",
    )


class GlobalDatasetPreparationParameters(BaseModel):
    """
    Parameters for preparing a dataset for training within the global configuration.
    Controls data splitting and filtering before being passed for the training.
    """

    subset_split: SubsetSplit = Field(
        title="Subset split",
        description="Configuration for splitting data into subsets",
    )
    filtering: Filtering = Field(title="Filtering", description="Configuration for filtering annotations")


class GlobalParameters(BaseModel):
    """
    Global parameters that are used within the application but are not directly passed to the training backend.
    These parameters still impact the final training outcome by controlling dataset preparation.
    """

    dataset_preparation: GlobalDatasetPreparationParameters = Field(
        title="Dataset preparation", description="Parameters for preparing the dataset"
    )


class TrainingConfiguration(BaseModel):
    """Configuration for model training"""

    task_id: str = Field(title="Task ID", description="Unique identifier for the task")
    model_config = ConfigDict(protected_namespaces=())  # avoid conflict with "model_" namespace
    model_manifest_id: str | None = Field(
        default=None,
        title="Model manifest ID",
        description="ID for the model manifest that defines the supported parameters and capabilities for training",
    )
    global_parameters: GlobalParameters = Field(
        title="Global parameters",
        description="Global configuration parameters for training",
    )
    hyperparameters: Hyperparameters = Field(title="Hyperparameters", description="Hyperparameters for training")
